#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/unistd.h>
#include <string.h>
#include <plugin.h>
#include <cleanup.h>
#include <signature.h>
#include <parser.h>

/* Snort rulefile parser
 * Genrally the parsing is done with turing state machines,
 * It isn't too hacky.
 *
 * TODO:
 * 	Allow variables to be used anywhere
 */

#define SNORT_MAXLEN 8192

proc_sig_rule signature_rule;
proc_sig_abort signature_abort;
proc_sig_criteria signature_criteria;
proc_sig_alert signature_alert;
proc_sig_commit signature_commit;

struct snort_element {
	struct snort_element *next;
	char *key;
	char *val;
	int neg;
};

struct snort_var {
	struct snort_var	*next;
	char			*name;
	char			*value;
};

/* Snort parsing variables */
char *snort_head[7];
int snort_negate[7];
struct snort_element *se_head;
struct snort_element *se_tail;
char *snort_targetops;
char *snort_msg;

/* Store variables */
struct snort_var *vars;

/* Snort keywords to ignore */
char *snort_hacks_ignore[]={
	"reference",
	"classtype",
	"sid",
	"rev",
	NULL
};

/* These are options for the "content" matcher */
char *snort_content_ops[]={
	"depth",
	"offset",
	"nocase",
	"regex",
	NULL
};

#define ELEN 128
char estr[ELEN];

struct snort_var *snort_var_find(char *key)
{
	struct snort_var *ret;

	for(ret=vars; ret; ret=ret->next) {
		if ( !strcmp(key, ret->name) )
			return ret;
	}
	
	return NULL;
}

int snort_addcrit(char *key, char *val, int negate)
{
	int i;
	struct snort_element *se;

	/* We need to ignore some stuff */
	for(i=0; snort_hacks_ignore[i]; i++) {
		if ( !strcmp(key, snort_hacks_ignore[i]) ) 
			return 1;
	}

	/* Content match options */
	for(i=0; snort_content_ops[i]; i++) {
		if ( !strcmp(key, snort_content_ops[i]) )
			return 1;
	}

	if ( !strcmp(key, "msg") ) {
		snort_msg=val;
		return 1;
	}

	if ( !strcmp(snort_head[0], "log") 
		&& !strcmp(key, "logto") ) {
		snort_targetops=val;
		return 1;
	}

	if ( !(se=malloc(sizeof(*se))) ) {
		return 0;
	}

	se->next=NULL;
	se->key=key;
	se->val=val;
	se->neg=negate;

	if ( !se_head ) {
		se_head=se;
		se_tail=se;
	}else{
		se_tail->next=se;
		se_tail=se;
	}
	
	return 1;
}

/* Parse the body of a snort rule (the bit after the bracket) */
#define zero_body() key=NULL; val=NULL; neg=0; state=0;
int snort_parse_body(char *line)
{
	char *key=NULL;
	char *val=NULL;
	int neg=0;

	char *cur;
	int state=0;
	int inq=0;
	
	/* Snort state machine parser */
	for(cur=line; ; cur++) {
		if ( *cur=='\n' || *cur=='\r' || *cur=='\0' ) break;

		if ( state==0 ) { /* Anywhere */
			if ( *cur!=' ' && *cur!='\t' ) {
				key=cur;
				state=1;
			}else if ( *cur==')' ) {
				*cur='\0';
				break;
			}
		}else if ( state==1 ) { /* Inside a key */
			if ( *cur==':' ) {
				*cur='\0';
				state=2;
			}else if ( *cur==';' && *(cur-1)!='\\' ) {
				*cur='\0';
				val=NULL;

				snort_addcrit(key, val, neg);
				zero_body();
			}
		}else if ( state==2 ) { /* Between key and value */
			if ( *cur=='!' ) {
				neg=1;
			}else if ( *cur!=' ' && *cur!='\t' ) {
				if ( *cur=='\"' ) {
					inq=1;
				}

				val=cur + inq;
				state=3;
			}
		}else if ( state==3 ) { /* In a value */
			if ( !inq && *cur=='\"' ) {
				inq=1;
			}else if ( inq && *cur=='\"' ) {
				inq=0;
			}else if ( !inq && *cur==';' && *(cur-1)!='\\' ) {

				if ( *(cur-1)=='\"' ) *(cur-1)='\0';
				*cur='\0';

				snort_addcrit(key, val, neg);
				zero_body();

			}else if ( !inq && *cur==')' ) {

				/* Allow missing end semicolon */
				if ( *(cur-1)=='\"' ) *(cur-1)='\0';
				*cur='\0';
				
				snort_addcrit(key, val, neg);
				zero_body();
			}
		}
	}
	
	return 1;
}

/* Actually commit a rule into the signature 
 * subsystem */
int snort_commit_rule(void)
{
	struct snort_element *s;
	struct rule *r;

	/* Create a new rule */
	if ( !(r=signature_rule(snort_head[1])) ) {
		snprintf(estr, ELEN, "Unable to create new '%s' rule",
			snort_head[1]);
		return 0;
	}

	/* Alert stuff */
	if ( !signature_alert(r, snort_head[0],
		snort_msg, snort_targetops) ) {
		snprintf(estr, ELEN, "Unable to set '%s' target",
			snort_head[0]);
		goto abort;
	}

	/* Source and destination addresses */
	if ( !signature_criteria(r, snort_negate[2],
			"src", snort_head[2]) ){
		snprintf(estr, ELEN, "Unable add head criteria to rule");
		goto abort;
	}

	if ( !signature_criteria(r, snort_negate[5],
			"dst", snort_head[5]) ){
		snprintf(estr, ELEN, "Unable add head criteria to rule");
		goto abort;
	}

	
	/* Ports */
	if ( strcmp(snort_head[3], "any") ) {
		if ( !signature_criteria(r, snort_negate[3],
				"sport", snort_head[3]) ) {
			snprintf(estr, ELEN, "Unable add head criteria to rule");
			goto abort;
		}
	}
	if ( strcmp(snort_head[6], "any") ) {
		if ( !signature_criteria(r, snort_negate[6],
				"dport", snort_head[6]) ) {
			snprintf(estr, ELEN, "Unable add head criteria to rule");
			goto abort;
		}
	}

	/* And all the options... */
	for(s=se_head; s; s=s->next) {
		if ( !signature_criteria(r, s->neg, s->key, s->val) ) {
			snprintf(estr, ELEN, "Unable add head criteria to rule");
			goto abort;
		}
	}

	/* lets go */
	if ( !signature_commit(r) ) {
		snprintf(estr, ELEN, "Rule failed to commit: %s",
			snort_msg);
		goto abort;
	}
	
	return 1;
abort:
	signature_abort(r);
	return 0;
}

int snort_var()
{
	struct snort_var *v, *k;
	size_t vlen=strlen(snort_head[2])+1;

	if ( (v=snort_var_find(snort_head[1])) ) {
		if ( !(v->value=realloc(v->value, vlen)) ) {
			snprintf(estr, ELEN, "Adding variable: %s",
				get_err());
			return 0;
		}

		/* Null termination is implicit */
		memcpy(v->value, snort_head[2], vlen);
		return 1;
	}else{
		if ( snort_head[2][0]=='$' ) {
			if ( !(k=snort_var_find(snort_head[2]+1)) ) {
				snprintf(estr, ELEN, "Unable to find $%s", 
					snort_head[2]+1);
				return 0;
			}

			snort_head[2]=k->value;
		}
		
		if ( !(v=malloc(sizeof(*v))) ) return 0;

		if ( !(v->name=strdup(snort_head[1])) ) {
			snprintf(estr, ELEN, "Updating variable: %s",
				get_err());
			free(v);
			return 0;
		}
		
		if ( !(v->value=strdup(snort_head[2])) ) {
			snprintf(estr, ELEN, "Updating variable: %s",
				get_err());
			free(v->name);
			free(v);
			return 0;
		}

		v->next=vars;
		vars=v;
		return 1;
	}
}

int snort_var_replace(char **ptr)
{
	struct snort_var *v;

	if ( **ptr=='$' ) {
		if ( !(v=snort_var_find(*ptr+1)) ) {
			return 0;
		}
		*ptr=v->value;
		return 1;
	}else return 1;
}

/* Snort up each line of the config  */
int snort_line(char *fn, int ln, char *line)
{
	struct snort_element *s;
	char *cur;
	int i;
	int end=0;
	int state=0;
	char any_sip[]="0/0";
	char any_dip[]="0/0";

	estr[0]=0;
	for(i=0; i<7; i++) {
		snort_negate[i]=0;
	}

	/* State machine parser, parse out first 7 elements
	 * seperated by any amount of whitespace ending in a '(' */
	for(i=0, cur=line; ; cur++) {
		if ( *cur=='\n' || *cur=='\r' || *cur=='\0' ){
			break;
		}

		if ( state == 0 ) {
			if ( *cur=='(' ) {
				if ( i==7 ) end=1;
				break;
			}
			
			if ( *cur=='!' ) {
				snort_negate[i]=1;
			}else if ( *cur!=' ' && 
				*cur!='\t' && *cur!='\0' ) {
				if ( i >= 7 ) break;

				snort_head[i]=cur;
				state=1;
				i++;
			}
		}else if ( state == 1 ) {
			if ( *cur==' ' || *cur=='\t' ) {
				*cur='\0';
				state=0;
			}
		}
	}
			
	if ( i==3 && !strcmp(snort_head[0], "var") ) {
		*cur=0;
		return snort_var();
	}

	/* Parse error */
	if ( !end ) {
		snprintf(estr, ELEN, "Incomplete line");
		return 0;
	}

	/* Its either "<-", "->" or "<>" */ 
	if ( strlen(snort_head[4]) !=2 ) {
		snprintf(estr, ELEN, "Invalid direction '%s'", snort_head[4]);
		return 0;
	}

	/* Parse the rest of the rule */
	se_head=NULL;
	se_tail=NULL;
	snort_msg=NULL;
	snort_targetops=NULL;
	snort_parse_body(++cur);

	snort_var_replace(&snort_head[2]);
	snort_var_replace(&snort_head[5]);

	/* any == "0/0" for IP addresses */
	if ( !strcmp(snort_head[2], "any") ) {
		snort_head[2]=any_sip;
	}

	if ( !strcmp(snort_head[5], "any") ){
		snort_head[5]=any_dip;
	}

	/* Go forwards */
	if ( snort_head[4][1]=='>' ) {
		if ( !snort_commit_rule() ) {
			return 0;
		}
	}

	/* Go backwards */
	if ( snort_head[4][0]=='<' ) {
		char *tadr, *tprt;

		/* Swap IPs and ports */
		tadr=snort_head[2];
		tprt=snort_head[3];
		snort_head[2]=snort_head[5];
		snort_head[3]=snort_head[6];
		snort_head[5]=tadr;
		snort_head[6]=tprt;

		if ( !snort_commit_rule() ) {
			return 0;
		}
	}

	for(s=se_head; s; ) {
		struct snort_element *tmp=s;
		s=s->next;
		free(tmp);	
	}
	
	return 1;
}

/* Parse the snort file, line by line */
int snort_run(char *fn)
{
	FILE *f;
	static char buf[SNORT_MAXLEN];
	int line=0;

	if ( !fn ) return 0;

	if ( (f=fopen(fn, "r"))==NULL ) {
		printf("snort: error: unable to open file '%s'\n", fn);
		return 0;
	}

	/* iterate each line in the file */
	while(fgets(buf, SNORT_MAXLEN, f)) {
		char *p=buf;
		
		line++;

		while ( *p && (*p=='\t' || *p==' ') ) p++;
		if ( *p=='\0' || *p=='\n' || *p=='\r' || *p=='#' ) continue;
		
		if ( !snort_line(fn, line, p) ) {
			printf("snort: %s:%i : %s\n", fn, line, estr);
			fclose(f);
			return 0;
		}
	}

	fclose(f);
	return 1;
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	proc_parser_add parser_add;
	int ok=0;
	
	plugin_check(in, out);
	
	PLUGIN_ID("parser.snort", "Snort ruleset files");
	PLUGIN_VERSION(2, 0);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");
	
	if ( !(parser_add=in->import("parser.add")) )
		return PLUGIN_ERR_OBJECT;
	
	if ( !(signature_rule=in->import("signature.rule")) )
		return PLUGIN_ERR_OBJECT;
	
	if ( !(signature_abort=in->import("signature.abort")) )
		return PLUGIN_ERR_OBJECT;
	
	if ( !(signature_criteria=in->import("signature.criteria")) )
		return PLUGIN_ERR_OBJECT;
	
	if ( !(signature_alert=in->import("signature.alert")) )
		return PLUGIN_ERR_OBJECT;

	if ( !(signature_commit=in->import("signature.commit")) )
		return PLUGIN_ERR_OBJECT;

	ok+=parser_add("snort", snort_run);
	
	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL;
}

int _PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
